/* spfquery - Sender Permitted From command line utility
 *	
 *  Author: Wayne Schlitt <wayne@midwestcs.com>
 *
 *  File:   spfquery.c
 *  Desc:   SPF command line utility
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of either:
 *
 *  a) the GNU General Public License as published by the Free Software
 *     Foundation; either version 1, or (at your option) any later
 *     version, or
 *
 *  OR
 *
 *  b) The following license:
 *
 *  License:
 *
 *  The libspf Software License, Version 1.0
 *
 *  Copyright (c) 2004 Wayne Schlitt  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 * 
 */

#include "spfquery.h"

void usage()
{
  fprintf( stderr, "Usage:\n" );
  fprintf( stderr, "\n" );
  fprintf( stderr, "spfquery [-v] [-f <file>|spf data options]\n" );

  fprintf( stderr, "\n" );

  fprintf( stderr, "spfquery -ipv4=<IP Address> -sender=<email address> -helo=<domain>\n" );
  fprintf( stderr, "spfquery -f test_data\n" );
  fprintf( stderr, "echo \"127.0.0.1 myname@mydomain.com helohost.com\" | spfquery -f -\n" );
}


int main( int argc, char *argv[] )
{
  int   c;
  int   res;

  char *opt_file          = NULL;

  char *opt_ipv4          = NULL;
  char *opt_sender        = NULL;
  char *opt_helo          = NULL;
  char *opt_rcpt_to       = NULL;

  char *opt_local         = NULL;
  char *opt_trusted       = NULL;
  char *opt_guess         = NULL;
  char *opt_exp           = NULL;
  char *opt_max_lookup    = NULL;
  char *opt_sanitize      = NULL;
  char *opt_debuglevel    = NULL;
  
  char *opt_myhostname    = "spfquery";

  char  in_line[4096];
  char  *p;

  int   dbg_level         = 0;

  peer_info_t *peer_info  = NULL; /* james */    
    
  FILE *fin;
    

  /*
   * check the arguments
   */

  while (1)
  {
    int option_index = 0;
    
    static struct option long_options[] =
    {
      { "file",                1,  0,  'f' },
      { "ipv4",                1,  0,  'i' },
      { "sender",              1,  0,  's' },
      { "helo",                1,  0,  'h' },
      { "rcpt-to",             1,  0,  'r' },
      { "local",               1,  0,  'l' },
      { "trusted",             1,  0,  't' },
      { "guess",               1,  0,  'g' },
      { "default_explanation", 1,  0,  'e' },
      { "max-lookup",          1,  0,  'm' },
      { "sanitize",            1,  0,  'c' },
      { "myhostname",          1,  0,  'n' },
      { "verbose",             1,  0,  'v' },
      { "help",                0,  0,  '?' },
      { 0,                     0,  0,  0   }
    };

    c = getopt_long_only (argc, argv, "f:i:s:h:r:v:ltgemcn",
			  long_options, &option_index);

    if (c == -1)
      break;

    switch (c)
    {
    case 'f':
      opt_file = optarg;
      break;
    case 'i':
      opt_ipv4 = optarg;
      break;
    case 's':
      opt_sender = optarg;
      break;
    case 'h':
      opt_helo = optarg;
      break;
    case 'r':
      opt_rcpt_to = optarg;
      /*fprintf( stderr, "Unimplemented option: -rcpt-to\n" );*/
      break;
    case 'l':
      opt_local = optarg;
      /*fprintf( stderr, "Unimplemented option: -rcpt-to\n" );*/
      break;
    case 't':
      opt_trusted = optarg;
      /*fprintf( stderr, "Unimplemented option: -rcpt-to\n" );*/
      break;
    case 'g':
      opt_guess = optarg;
      /*fprintf( stderr, "Unimplemented option: -rcpt-to\n" );*/
      break;
    case 'e':
      opt_exp = optarg;
      /*fprintf( stderr, "Unimplemented option: -rcpt-to\n" );*/
      break;
    case 'm':
      opt_max_lookup = optarg;
      /*fprintf( stderr, "Unimplemented option: -rcpt-to\n" );*/
      break;
    case 'c':			/* "clean"			*/
      opt_sanitize = optarg;
      /*fprintf( stderr, "Unimplemented option: -rcpt-to\n" );*/
      break;
    case 'n':			/* "name"			*/
      opt_myhostname = optarg;
      /*fprintf( stderr, "Unimplemented option: -rcpt-to\n" );*/
      break;
    case 0:
    case '?':
      usage();
      return 1;
      break;
    case 'v':
      opt_debuglevel = optarg;
      dbg_level = atoi(opt_debuglevel);
      confg.level = dbg_level;
      printf("DEBUGGING LEVEL IS: %i\n", dbg_level);
      /*dbg_level++;*/
      break;
    default:
      fprintf( stderr, "Error: getopt returned character code 0%o ??\n", c);
    }
  }

  if (optind != argc)
  {
    usage();
    return 1;
  }

  /*
   * process the SPF request
   */

  if (opt_ipv4 == NULL || opt_sender == NULL || opt_helo == NULL) 
  {
    if (opt_file == NULL ||
	opt_ipv4 != NULL || opt_sender != NULL || opt_helo != NULL)
    {
      usage();
      return 1;
    }

    /*
     * the requests are on STDIN
     */
	
    if (strcmp( opt_file, "-" ) == 0) 
      fin = stdin;
    else
      fin = fopen( opt_file, "r" );

    if (!fin) 
    {
      fprintf( stderr, "Could not open: %s\n", opt_file );
      return 1;
    }
	
    while ( TRUE )
    {
      if( fgets( in_line, sizeof( in_line ), stdin ) == NULL )
	      break;
        
      p = strchr( in_line, '\n' );
      
      if( p != NULL )
	      *p = '\0';
        
      p = in_line;
      
      while( (opt_ipv4 = strsep( &p, " \t\n" )) != NULL
	     && *opt_ipv4 == '\0' )
	    ;
      
      if (opt_ipv4 == NULL)
	      break;
        
      while( (opt_sender = strsep( &p, " \t\n" )) != NULL
	     && *opt_ipv4 == '\0' )
	    ;
      
      while( (opt_helo = strsep( &p, " \t\n" )) != NULL
	     && *opt_ipv4 == '\0' )
	    ;
      
      if ( dbg_level ) 
      {
	      printf ("ipv4:   %s\n", opt_ipv4);
	      printf ("sender: %s\n", opt_sender);
	      printf ("helo:   %s\n", opt_helo);
      }
      
      if (peer_info == NULL)
	      peer_info = SPF_init(opt_myhostname, opt_ipv4, NULL, NULL, NULL,
          FALSE, FALSE);
        
      if (opt_helo != NULL)
	      SPF_smtp_helo(peer_info, opt_helo);
        
      if (opt_sender != NULL)
	      SPF_smtp_from(peer_info, opt_sender);
      
      peer_info->RES = SPF_policy_main(peer_info);
      res = peer_info->RES;
      
      printf( "%s\n%s\n%s\n",
	      peer_info->rs, peer_info->error, peer_info->explain);

      SPF_close(peer_info);
        
    }
  }
  else
  {
    if (opt_file != NULL)
    {
      usage();
      return 1;
    }

    /*
     * The request was given on the command line
     */
    
    if ( dbg_level ) 
    {
      printf ("ipv4:   %s\n", opt_ipv4);
      printf ("sender: %s\n", opt_sender);
      printf ("helo:   %s\n", opt_helo);
    }
	    
    if (peer_info == NULL)
      peer_info = SPF_init(opt_myhostname, opt_ipv4, NULL, NULL, NULL,
        FALSE, FALSE);
        
    if (opt_helo != NULL)
      SPF_smtp_helo(peer_info, opt_helo);
        
    if (opt_sender != NULL)
      SPF_smtp_from(peer_info, opt_sender);
	
    peer_info->RES = SPF_policy_main(peer_info);

    printf( "%s\n%s\n%s\n",
	    peer_info->rs, peer_info->error, peer_info->explain);
	
    res = peer_info->RES;
    
    SPF_close(peer_info);

    return res;
  }
    
  return 0;
}
